"""
Training graph network property models
"""
import logging
import os
from glob import glob
from typing import List, Optional, Union

import numpy as np
import tensorflow as tf
from ase import Atoms
from pymatgen.core import Molecule, Structure

from m3gnet.callbacks import ManualStop
from m3gnet.graph import MaterialGraph, MaterialGraphBatch
from m3gnet.layers import AtomRef
from m3gnet.models import M3GNet

from ._metrics import MONITOR_MAPPING, _get_metric, _get_metric_string

logging.basicConfig()
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class Trainer:
    """
    Trainer for material properties
    """

    def __init__(self, model: M3GNet, optimizer: tf.keras.optimizers.Optimizer):
        """
        Args:
            model (M3GNet): a M3GNet model
            optimizer (tf.keras.optimizers.Optimizer): a keras optimizer
        """
        self.model = model
        self.optimizer = optimizer
        self.initial_epoch = 0

    def restart_from_directory(self, dirname: str):
        """
        Continue previous model training from a directory
        Args:
            dirname (str): directory name
        Returns:
        """
        filenames = glob(os.path.join(dirname, "*.index"))
        latest_model = sorted(filenames, key=os.path.getctime)[-1]
        latest_model = latest_model.rsplit(".", 1)[0]
        self.initial_epoch = int(latest_model.split("/", 1)[1].split("-", 1)[0]) + 1
        self.model.load_weights(latest_model)

    def train(
        self,
        graphs_or_structures: List,
        targets: List,
        validation_graphs_or_structures: List = None,
        validation_targets: List = None,
        loss: tf.keras.losses.Loss = tf.keras.losses.MSE,
        train_metrics: Optional[List] = None,
        val_metrics: Optional[List] = None,
        val_monitor: str = "val_mae",
        batch_size: int = 128,
        epochs: int = 1000,
        callbacks: List = None,
        save_checkpoint: bool = True,
        verbose: int = 1,
        clip_norm: Optional[float] = 10.0,
        fit_per_element_offset: bool = False,
    ):
        """
        Args:
            graphs_or_structures (list): a list of MaterialGraph or structures
            targets (list): list of properties in float
            validation_graphs_or_structures (list): optional list of validation
                graphs or structures
            validation_targets (list): optional list of properties
            loss (tf.keras.losses.Loss): loss object
            train_metrics (list): list of train metrics
            val_metrics (list): list of validation metrics
            val_monitor (str): field to monitor during validation, e.g.,
                "val_mae", "val_acc" or "val_auc"
            batch_size (int): batch size of combining graphs
            epochs (int): epochs for training the data
            callbacks (list): list of callback functions
            save_checkpoint (bool): whether to save model check point
            verbose (bool): whether to show model training progress
            clip_norm (float): gradient norm clip
            fit_per_element_offset (bool): whether to train an element-wise
                offset, e.g., elemental energies etc. If trained, such energy
                will be summed to the neural network predictions.

        Returns: None
        """

        if isinstance(graphs_or_structures[0], MaterialGraph):
            graphs = graphs_or_structures
        elif isinstance(graphs_or_structures[0], (Structure, Molecule, Atoms)):
            graphs = [self.model.graph_converter(i) for i in graphs_or_structures]
        else:
            raise ValueError("Graph types not recognized")

        if fit_per_element_offset:
            ar = AtomRef(max_z=self.model.n_atom_types + 1)
            ar.fit(graphs, targets)
            self.model.set_element_refs(ar.property_per_element)

        val_metrics = val_metrics or ["mae"]
        mgb = MaterialGraphBatch(graphs, targets, batch_size=batch_size)

        if train_metrics is not None:
            train_metrics = [_get_metric(metric) for metric in train_metrics]
            train_metric_names = [f"{_get_metric_string(i) }" for i in train_metrics]
            has_train_metrics = True
        else:
            train_metric_names = []
            has_train_metrics = False

        if val_metrics is not None:
            val_metrics = [_get_metric(metric) for metric in val_metrics]
            val_metric_names = [f"val_{_get_metric_string(i)}" for i in val_metrics]
            has_val_metrics = True
        else:
            val_metric_names = []
            has_val_metrics = False

        if validation_graphs_or_structures is not None and validation_targets is not None:
            has_validation = True
            if isinstance(validation_graphs_or_structures[0], MaterialGraph):
                validation_graphs = validation_graphs_or_structures
            elif isinstance(validation_graphs_or_structures[0], (Structure, Molecule, Atoms)):
                validation_graphs = [self.model.graph_converter(i) for i in validation_graphs_or_structures]
            else:
                raise ValueError("Graph types not recognized")

            mgb_val = MaterialGraphBatch(
                validation_graphs,
                validation_targets,
                batch_size=batch_size,
                shuffle=False,
            )
        else:
            has_validation = False

        if callbacks is None:
            callbacks = [tf.keras.callbacks.History()]

        callbacks.append(ManualStop())

        if has_validation and save_checkpoint:
            if val_monitor not in val_metric_names:
                raise ValueError(f"val_monitor {val_monitor} not in the " f"val_metric_names {val_metric_names}")

            callbacks.append(
                tf.keras.callbacks.ModelCheckpoint(
                    filepath=f"callbacks/{{epoch:05d}}-{{{val_monitor}:.6f}}",
                    monitor=val_monitor,
                    save_weights_only=True,
                    save_best_only=True,
                    mode=MONITOR_MAPPING[val_monitor],
                )
            )

        if verbose:
            pbar = tf.keras.callbacks.ProgbarLogger(count_mode="steps")
            pbar.set_params({"verbose": verbose, "epochs": epochs})
            callbacks.append(pbar)

        if len(callbacks) > len(set(i.__class__ for i in callbacks)):
            logger.warning("Duplicated callbacks found")

        callback_list = tf.keras.callbacks.CallbackList(callbacks)
        callback_list.on_train_begin()
        callback_list.set_model(self.model)
        stop_training = False

        @tf.function(experimental_relax_shapes=True)
        def predict(model, graph_batch):
            return model(graph_batch)

        @tf.function(experimental_relax_shapes=True)
        def train_one_step(model, graph_list, target_list):
            with tf.GradientTape() as tape:
                pred_list: tf.Tensor = model(graph_list)
                loss_val = loss(target_list, pred_list)
            grads = tape.gradient(loss_val, model.trainable_variables)
            return loss_val, grads, pred_list

        for epoch in range(self.initial_epoch, epochs):
            callback_list.on_epoch_begin(epoch=epoch, logs={"epoch": epoch})
            epoch_loss_avg = tf.keras.metrics.Mean()

            if has_val_metrics:
                val_metric_values = {}
            if has_train_metrics:
                train_metric_values = {}

            train_predictions: Union[list, tf.Tensor] = []
            train_targets: Union[list, tf.Tensor] = []

            for batch_index, batch in enumerate(mgb):
                callback_list.on_batch_begin(batch=batch_index)
                graph_batch, target_batch = batch
                if isinstance(graph_batch, MaterialGraph):
                    graph_batch = graph_batch.as_list()
                if isinstance(target_batch, np.ndarray) and target_batch.ndim == 1:
                    target_batch = target_batch.reshape((-1, 1))

                loss_val, grads, predictions = train_one_step(self.model, graph_batch, target_batch)
                global_norm = tf.linalg.global_norm(grads)

                if clip_norm is not None:
                    grads, _ = tf.clip_by_global_norm(grads, clip_norm, use_norm=global_norm)

                self.optimizer.apply_gradients(zip(grads, self.model.trainable_variables))

                epoch_loss_avg.update_state(loss_val)
                logs = {"loss": epoch_loss_avg.result().numpy()}

                callback_list.on_batch_end(batch=batch_index, logs=logs)
                train_predictions.append(predictions)
                train_targets.append(target_batch)

                if getattr(self.model, "stop_training", False):
                    stop_training = True
                    break

            if stop_training:
                break

            epoch_log = {"loss": epoch_loss_avg.result().numpy()}
            train_predictions = tf.concat(train_predictions, axis=0)
            train_targets = tf.concat(train_targets, axis=0)
            if has_train_metrics:
                for index, metric_name in enumerate(train_metric_names):
                    train_metric_values[metric_name] = train_metrics[index](  # type: ignore
                        train_targets.numpy().ravel(), train_predictions.numpy().ravel()
                    )
                epoch_log.update(
                    **{metric_name: train_metric_values[metric_name].numpy() for metric_name in train_metric_names}
                )

            if has_validation:
                val_predictions = []
                val_targets = []
                for val_index, batch in enumerate(mgb_val):
                    graph_batch, target_batch = batch
                    if isinstance(graph_batch, MaterialGraph):
                        graph_batch = graph_batch.as_list()
                    if target_batch.ndim == 1:
                        target_batch = target_batch.reshape((-1, 1))
                    predictions = predict(self.model, graph_batch)
                    val_predictions.append(predictions)
                    val_targets.append(target_batch)
                val_predictions = tf.concat(val_predictions, axis=0)
                val_targets = tf.concat(val_targets, axis=0)
                if has_val_metrics:
                    for index, metric_name in enumerate(val_metric_names):
                        val_metric_values[metric_name] = val_metrics[index](
                            val_targets.numpy().ravel(),  # type: ignore
                            val_predictions.numpy().ravel(),  # type: ignore
                        )
                    val_logs = {metric_name: val_metric_values[metric_name].numpy() for metric_name in val_metric_names}
                    epoch_log.update(**val_logs)
            callback_list.on_epoch_end(epoch=epoch, logs=epoch_log)
            mgb.on_epoch_end()
        callback_list.on_train_end()

        if save_checkpoint and has_validation:
            best_model_weights = sorted(glob("callbacks/*.index"), key=os.path.getctime)[-1]
            best_model_weights = best_model_weights.rsplit(".", 1)[0]
            self.model.load_weights(best_model_weights)
            self.model.save("best_model")



"""
Installation for m3gnet
"""

import os
import re

from setuptools import find_packages
from setuptools import setup, Extension
import numpy as np
from Cython.Build import cythonize

this_dir = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_dir, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open("m3gnet/__init__.py", encoding="utf-8") as fd:
    for line in fd.readlines():
        m = re.search('__version__ = "(.*)"', line)
        if m:
            version = m.group(1)
            break

extension = [
    Extension(
        "m3gnet.graph._threebody_indices",
        ["m3gnet/graph/_threebody_indices.pyx"],
    )
]

setup(
    name="m3gnet",
    version=version,
    install_requires=[
        "pymatgen>=2022.1.1",
        "tensorflow",
        "numpy",
        "monty",
        "sympy",
        "cython",
        "ase",
    ],
    description="Materials Graph with Three-body Interactions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Chi Chen",
    author_email="chen08013@gmail.com",
    download_url="https://github.com/materialsvirtuallab/m3gnet",
    license="BSD",
    extras_require={
        "model_saving": ["h5py"],
        "tensorflow with gpu": ["tensorflow-gpu>=2.7"],
    },
    packages=find_packages(),
    package_data={
        "m3gnet": ["*.json", "*.md"],
    },
    include_package_data=True,
    ext_modules=cythonize(extension),
    include_dirs=np.get_include(),
    keywords=[
        "materials",
        "science",
        "machine",
        "learning",
        "deep",
        "graph",
        "networks",
        "neural",
        "force field",
        "interatomic potential",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# -*- coding: utf-8 -*-
import unittest

import numpy as np
import tensorflow as tf

from m3gnet.utils import (
    broadcast_states_to_atoms,
    broadcast_states_to_bonds,
    get_length,
    get_range_indices_from_n,
    get_segment_indices_from_n,
    register,
    register_plain,
    repeat_with_n,
    unsorted_segment_fraction,
    unsorted_segment_softmax,
)


class TestTF(unittest.TestCase):
    def test_calculations(self):
        np.testing.assert_array_almost_equal(get_segment_indices_from_n([2, 3]), [0, 0, 1, 1, 1])

        np.testing.assert_array_almost_equal(get_range_indices_from_n([2, 3]), [0, 1, 0, 1, 2])

        np.testing.assert_array_almost_equal(
            repeat_with_n([[0, 0], [1, 1], [2, 2]], [1, 2, 3]),
            [[0, 0], [1, 1], [1, 1], [2, 2], [2, 2], [2, 2]],
        )

        x = np.random.normal(size=(10, 3))
        np.testing.assert_array_almost_equal(get_length(x), np.linalg.norm(x, axis=1))

    def test_segments(self):
        x = np.array([1.0, 1.0, 2.0, 3.0])
        res = unsorted_segment_fraction(x, [0, 0, 1, 1], 2)
        np.testing.assert_array_almost_equal(res, [0.5, 0.5, 0.4, 0.6])

        res = unsorted_segment_softmax(x, [0, 0, 1, 1], 2)
        np.testing.assert_array_almost_equal(res, [0.5, 0.5, 0.26894143, 0.7310586])

    def test_broadcast(self):
        from m3gnet.graph import Index

        graph = [[] for i in range(11)]
        graph[Index.STATES] = np.array([[0.0, 0.0], [1.0, 1.0]])
        graph[Index.N_BONDS] = np.array([5, 6])
        self.assertTupleEqual(tuple(broadcast_states_to_bonds(graph).shape), (11, 2))

        graph[Index.N_ATOMS] = np.array([3, 5])
        self.assertTupleEqual(tuple(broadcast_states_to_atoms(graph).shape), (8, 2))

    def test_register(self):
        class TestClass(tf.keras.layers.Layer):
            pass

        register(TestClass)
        self.assertTrue("m3gnet>TestClass" in tf.keras.utils.get_custom_objects())

        class TestClass2(tf.keras.layers.Layer):
            pass

        register_plain(TestClass2)

        self.assertTrue("TestClass2" in tf.keras.utils.get_custom_objects())


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest

import numpy as np
from pymatgen.core import Lattice, Structure

from m3gnet.layers import AtomRef


class TestRef(unittest.TestCase):
    def test_ref(self):
        ar = AtomRef(property_per_element=np.random.normal(size=(94,)))
        s1 = Structure(Lattice.cubic(3), ["Mo", "Mo"], [[0, 0, 0], [0.5, 0.5, 0.5]])
        s2 = s1 * [1, 2, 1]
        ar.fit([s1, s2], [-10, -20])
        self.assertAlmostEqual(ar.property_per_element[42], -5.0)
        new_energies = ar.transform([s1, s2], [-10, -20])
        self.assertTrue(np.allclose(new_energies, [0, 0]))
        energies = ar.inverse_transform([s1, s2], [0, 0])
        self.assertTrue(np.allclose(energies, [-10, -20]))
        self.assertAlmostEqual(ar.predict_properties(s1 * 2), -80)


if __name__ == "__main__":
    unittest.main()



"""
Graph package
"""

from ._batch import (
    MaterialGraphBatch,
    MaterialGraphBatchEnergyForceStress,
    assemble_material_graph,
)
from ._compute import (
    get_pair_vector_from_graph,
    include_threebody_indices,
    tf_compute_distance_angle,
)
from ._converters import BaseGraphConverter, RadiusCutoffGraphConverter
from ._structure import get_fixed_radius_bonding
from ._types import Index, MaterialGraph

__all__ = [
    "MaterialGraph",
    "MaterialGraphBatch",
    "MaterialGraphBatchEnergyForceStress",
    "assemble_material_graph",
    "BaseGraphConverter",
    "RadiusCutoffGraphConverter",
    "get_fixed_radius_bonding",
    "include_threebody_indices",
    "Index",
    "get_pair_vector_from_graph",
    "tf_compute_distance_angle",
]



# -*- coding: utf-8 -*-
import json
import os
import unittest

import numpy as np

from m3gnet.graph import MaterialGraph
from m3gnet.layers import MultiFieldReadout, ReduceReadOut, Set2Set

CWD = os.path.abspath(os.path.dirname(__file__))


class TestReadout(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        with open(os.path.join(CWD, "data", "test_NaCl.json")) as f:
            g_list = json.load(f)
        cls.graph = MaterialGraph.from_list(g_list)
        cls.graph_w_atom_features = cls.graph.copy()
        cls.graph_w_atom_features.atoms = np.random.normal(size=(cls.graph.atoms.shape[0], 10))

    def test_set2set(self):
        set2set = Set2Set(units=3, num_steps=2, field="bonds")
        res = set2set(self.graph.as_list())
        self.assertTrue(res.shape == (1, 6))

    def test_atomreduce(self):
        ar = ReduceReadOut(method="mean", field="atoms")
        result = ar(self.graph.as_list())
        self.assertTrue(result.shape == (1, 1))

    def test_multifield(self):
        atom = ReduceReadOut(method="mean", field="atoms")
        bond = Set2Set(units=3, num_steps=2, field="bonds")
        multi = MultiFieldReadout(bond_readout=bond, atom_readout=atom)
        graph = self.graph.copy()
        graph.atoms = np.zeros_like(graph.atoms, dtype=np.float32)
        x = multi(graph.as_list())
        self.assertTrue(x.shape == (1, 7))


if __name__ == "__main__":
    unittest.main()



"""
Collate material graphs
"""
from typing import AnyStr, List, Optional, Tuple, Union, overload

import numpy as np
import tensorflow as tf
from tensorflow.keras.utils import Sequence

from ._types import Index, MaterialGraph


class MaterialGraphBatch(Sequence):
    """
    MaterialGraphBatch for generating batches of MaterialGraphs
    """

    def __init__(
        self,
        graphs: List[MaterialGraph],
        targets: Optional[np.ndarray] = None,
        batch_size: int = 128,
        shuffle: bool = True,
    ):
        """
        Args:
            graphs (list): list of MaterialGraph or MaskedMaterialGraph
            targets (np.ndarray): array of targets
            batch_size (int): batch size
            shuffle (bool): whether to shuffle graphs at the end of each epoch
        """
        self.graphs = graphs
        self.targets = np.array(targets, dtype="float32")
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.total_n = len(graphs)
        self.max_step = int(np.ceil(self.total_n / batch_size))
        self.graph_index = np.arange(self.total_n)
        self._current_step = 0
        if self.shuffle:
            self.graph_index = np.random.permutation(self.graph_index)

    def __getitem__(self, index) -> Union[MaterialGraph, Tuple[MaterialGraph, np.ndarray]]:
        """
        Get the index-th batch of data. Returns a MaterialGraph object that
        contains many structures

        Args:
            index: int, the batch index

        Returns: MaterialGraph

        """
        graph_indices = self.graph_index[index * self.batch_size : (index + 1) * self.batch_size]
        graphs = [self.graphs[i] for i in graph_indices]
        new_graph: MaterialGraph = assemble_material_graph(graphs)
        targets: np.ndarray = None if self.targets is None else self.targets[graph_indices]
        if targets is None:
            return new_graph
        return new_graph, targets

    def __len__(self) -> int:
        return self.max_step

    def on_epoch_end(self) -> None:
        """
        Operations at the end of each training epoch
        """
        if self.shuffle:
            self.graph_index = np.random.permutation(self.graph_index)


class MaterialGraphBatchEnergyForceStress(MaterialGraphBatch):
    """
    MaterialGraphBatch for generating batches of MaterialGraphs
    """

    def __init__(
        self,
        graphs: List[MaterialGraph],
        energies: Union[List, np.ndarray] = None,
        forces: List[np.ndarray] = None,
        stresses: List[np.ndarray] = None,
        batch_size: int = 128,
        shuffle: bool = True,
    ):
        """
        Args:
            graphs (list): list of MaterialGraph or MaskedMaterialGraph
            energies (np.ndarray): array of targets
            forces (List): list of force matrix
            stresses (List or np.ndarray): list of stress matrix
            batch_size (int): batch size
            shuffle (bool): whether to shuffle graphs at the end of each epoch
        """
        super().__init__(graphs=graphs, targets=energies, batch_size=batch_size, shuffle=shuffle)
        self.forces = forces
        self.stresses = np.array(stresses, dtype="float32") if stresses is not None else None

    def __getitem__(self, index):
        graph_indices = self.graph_index[index * self.batch_size : (index + 1) * self.batch_size]
        graphs, energies = super().__getitem__(index)
        forces = np.concatenate([self.forces[i] for i in graph_indices], axis=0)
        forces = np.array(forces, dtype="float32")
        return_values = [graphs, (energies, forces)]
        if self.stresses is not None:
            stresses = np.array([self.stresses[i] for i in graph_indices])
            return_values[1] += (stresses,)
        return return_values


def _check_none_field(graph_list, field) -> bool:
    if any(getattr(g, field, None) is None for g in graph_list):
        return True
    return False


def _concatenate(list_of_arrays: List, name: AnyStr) -> Optional[np.ndarray]:
    """
    Concatenate list of array on the first dimension
    Args:
        list_of_arrays (List): list of arrays
        name (String): Used for error message

    Returns: concatenated array
    """
    num_none = sum([i is None for i in list_of_arrays])

    if num_none == len(list_of_arrays):
        return None

    if num_none == 0:
        if isinstance(list_of_arrays[0], tf.Tensor):
            return tf.concat(list_of_arrays, axis=0)
        return np.concatenate(list_of_arrays, axis=0)

    none_indices = [i for i, j in enumerate(list_of_arrays) if j is None]
    raise ValueError(
        f"The {name!r} properties of the graph indices {str(none_indices)!r} "
        f"are None while the rest graphs are not None. Hence the graphs "
        f"cannot be assembled."
    )


@overload
def assemble_material_graph(graphs: List[MaterialGraph]) -> MaterialGraph:
    ...


@overload
def assemble_material_graph(graphs: List[List]) -> List:
    ...


def assemble_material_graph(graphs):
    """
    Collate a list of MaterialGraph and form a single MaterialGraph

    Args:
        graphs: list of MaterialGraph

    Returns: a single MaterialGraph
    """
    if isinstance(graphs[0], List):
        return _assemble_material_graph_list(graphs)
    atoms = _concatenate([i.atoms for i in graphs], "atoms")
    bonds = _concatenate([i.bonds for i in graphs], "bonds")
    states = _concatenate([i.states for i in graphs], "states")
    bond_atom_indices = _concatenate([i.bond_atom_indices for i in graphs], "bond_atom_indices")

    n_atoms = _concatenate([i.n_atoms for i in graphs], "n_atoms")
    n_bonds = _concatenate([i.n_bonds for i in graphs], "n_bonds")

    n_atom_cumsum = np.cumsum([0] + [i.n_atom for i in graphs[:-1]])
    n_bond_every = [i.n_bond for i in graphs]
    bond_atom_indices += np.repeat(n_atom_cumsum, n_bond_every)[:, None]

    atom_positions = _concatenate([i.atom_positions for i in graphs], "atom_positions")
    bond_weights = _concatenate([i.bond_weights for i in graphs], "bond_weights")
    pbc_offsets = _concatenate([i.pbc_offsets for i in graphs], "pbc_offsets")
    lattices = _concatenate([i.lattices for i in graphs], "lattices")
    if graphs[0].has_threebody:
        triple_bond_indices = _concatenate([i.triple_bond_indices for i in graphs], "triple_bond_indices")
        n_bond_cumsum = np.cumsum([0] + n_bond_every[:-1])
        n_triple_s = _concatenate([i.n_triple_s for i in graphs], "n_triple_s")
        n_triple_s_temp = np.array([sum(i.n_triple_s) for i in graphs])
        triple_bond_indices += np.repeat(n_bond_cumsum, n_triple_s_temp)[:, None]
        triple_bond_lengths = _concatenate([i.triple_bond_lengths for i in graphs], "triple_bond_lengths")
        theta = _concatenate([i.theta for i in graphs], "theta")
        phi = _concatenate([i.phi for i in graphs], "phi")
        n_triple_ij = _concatenate([i.n_triple_ij for i in graphs], "n_triple_ij")
        n_triple_i = _concatenate([i.n_triple_i for i in graphs], "n_triple_i")

    else:
        triple_bond_indices = None
        triple_bond_lengths = None
        theta = None
        phi = None
        n_triple_ij = None
        n_triple_i = None
        n_triple_s = None

    return MaterialGraph(
        atoms=atoms,
        bonds=bonds,
        states=states,
        bond_atom_indices=bond_atom_indices,
        n_atoms=n_atoms,
        n_bonds=n_bonds,
        atom_positions=atom_positions,
        bond_weights=bond_weights,
        pbc_offsets=pbc_offsets,
        lattices=lattices,
        triple_bond_indices=triple_bond_indices,
        triple_bond_lengths=triple_bond_lengths,
        theta=theta,
        phi=phi,
        n_triple_ij=n_triple_ij,
        n_triple_i=n_triple_i,
        n_triple_s=n_triple_s,
    )


def _assemble_material_graph_list(graphs: List[List]) -> List:
    """
    Collate a list of MaterialGraph and form a single MaterialGraph

    Args:
        graphs: list of MaterialGraph

    Returns: a single MaterialGraph
    """

    graph = [None] * len(graphs[0])

    for i in [
        "atoms",
        "bonds",
        "states",
        "bond_atom_indices",
        "n_atoms",
        "n_bonds",
        "atom_positions",
        "bond_weights",
        "pbc_offsets",
        "lattices",
    ]:
        ind = getattr(Index, i.upper())
        graph[ind] = _concatenate([g[ind] for g in graphs], i)
    n_atoms = np.concatenate([i[Index.N_ATOMS] for i in graphs[:-1]])
    n_atom_cumsum = np.cumsum(np.concatenate([[0], n_atoms]))
    n_bond_every = [i[Index.BONDS].shape[0] for i in graphs]
    graph[Index.BOND_ATOM_INDICES] += np.repeat(n_atom_cumsum, n_bond_every)[:, None]

    if graphs[0][Index.N_TRIPLE_IJ] is not None:
        for i in [
            "triple_bond_indices",
            "n_triple_s",
            "triple_bond_lengths",
            "theta",
            "phi",
            "n_triple_ij",
            "n_triple_i",
        ]:
            ind = getattr(Index, i.upper())
            graph[ind] = _concatenate([g[ind] for g in graphs], i)
        n_bond_cumsum = np.cumsum([0] + n_bond_every[:-1])
        n_triple = np.array([sum(graph[Index.N_TRIPLE_S]) for graph in graphs])
        graph[Index.TRIPLE_BOND_INDICES] += np.repeat(n_bond_cumsum, n_triple)[:, None]
    return graph



"""
Readout compress a graph into a vector
"""
from typing import List, Optional

import tensorflow as tf

from m3gnet.config import DataType
from m3gnet.graph import Index
from m3gnet.utils import (
    get_segment_indices_from_n,
    register,
    unsorted_segment_fraction,
    unsorted_segment_softmax,
)

from ._core import METHOD_MAPPING, MLP


@register
class ReadOut(tf.keras.layers.Layer):
    """
    Readout reduces a graph into a tensor
    """

    def call(self, graph: List, **kwargs) -> tf.Tensor:
        """
        Args:
            graph (list): list repr of a MaterialGraph
            **kwargs:
        Returns: tf.Tensor, tensor readout

        """
        raise NotImplementedError


@register
class MultiFieldReadout(ReadOut):
    """
    Read both bond and atom
    """

    def __init__(
        self,
        bond_readout: Optional[ReadOut] = None,
        atom_readout: Optional[ReadOut] = None,
        include_states: bool = True,
        merge: Optional[tf.keras.layers.Layer] = None,
        **kwargs,
    ):
        """

        Args:
            bond_readout (ReadOut): the bond readout instance
            atom_readout (ReadOut: the atom readout instance
            include_states (bool): whether to include states
            merge (tf.keras.layers.Layer): method to merge different readout
            **kwargs:
        """
        self.bond_readout = bond_readout
        self.atom_readout = atom_readout
        self.include_states = include_states
        self.merge = merge or tf.keras.layers.Concatenate(axis=-1)
        super().__init__(**kwargs)

    def call(self, graph: List, **kwargs) -> tf.Tensor:
        """
        Args:
            graph (list): list repr of a MaterialGraph
            **kwargs:
        Returns:

        """
        features = []
        if self.bond_readout is not None:
            features.append(self.bond_readout(graph))

        if self.atom_readout is not None:
            features.append(self.atom_readout(graph))

        if self.include_states and graph[Index.STATES] is not None:
            features.append(graph[Index.STATES])

        if len(features) > 1:
            return self.merge(features)
        return features[0]

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update(
            {
                "bond_readout": self.bond_readout,
                "atom_readout": self.atom_readout,
                "include_states": self.include_states,
                "merge": self.merge,
            }
        )
        return config


@register
class ReduceReadOut(ReadOut):
    """
    Reduce atom or bond attributes into lower dimensional tensors as readout.
    This could be summing up the atoms or bonds, or taking the mean, etc.
    """

    def __init__(self, method: str = "mean", field="atoms", **kwargs):
        """
        Args:
            method (str): method for the reduction
            field (str): the field of MaterialGraph to perform the reduction
            **kwargs:
        """
        self.method = method
        self.field = field
        super().__init__(**kwargs)
        self.method_func = METHOD_MAPPING.get(method)

    def call(self, graph: List, **kwargs) -> tf.Tensor:
        """
        Args:
            graph (list): list repr of a MaterialGraph
            **kwargs:
        Returns:

        """
        field = graph[getattr(Index, self.field.upper())]
        n_field = graph[getattr(Index, f"n_{self.field}".upper())]
        return self.method_func(  # type: ignore
            field,
            get_segment_indices_from_n(n_field),
            num_segments=tf.shape(n_field)[0],
        )

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update({"method": self.method, "field": self.field})
        return config


@register
class Set2Set(ReadOut):
    """
    The Set2Set readout function
    """

    def __init__(self, units: int, num_steps: int, field: str, **kwargs):
        """
        Args:
            units (int): number of neurons in the set2set layer
            num_steps (int): number of LSTM steps
            field (str): the field of MaterialGraph to perform the readout
            **kwargs:
        """

        super().__init__(**kwargs)
        self.units = units
        self.num_steps = num_steps
        self.lstm = tf.keras.layers.LSTM(units=units, stateful=False, return_state=True)
        self.dense = tf.keras.layers.Dense(units=units)
        self.field = field

    def call(self, graph: List, **kwargs) -> tf.Tensor:
        """
        Args:
            graph (list): list repr of a MaterialGraph
            **kwargs:
        Returns:

        """
        field = graph[getattr(Index, self.field.upper())]
        field = self.dense(field)
        n_struct = tf.shape(graph[Index.N_ATOMS])[0]
        counts = graph[getattr(Index, f"n_{self.field}".upper())]
        indices = get_segment_indices_from_n(counts)
        q_star = tf.zeros([n_struct, 2 * self.units], dtype=DataType.tf_float)
        h = [
            tf.zeros([n_struct, self.units], dtype=DataType.tf_float),
            tf.zeros([n_struct, self.units], dtype=DataType.tf_float),
        ]
        for _ in range(self.num_steps):
            qt, state_h, state_c = self.lstm(tf.expand_dims(q_star, axis=1), initial_state=h)
            h = [state_h, state_c]
            eit = field * tf.repeat(qt, repeats=counts, axis=0)
            ait = unsorted_segment_softmax(eit, indices, num_segments=n_struct)
            rt = tf.math.unsorted_segment_sum(ait * field, indices, num_segments=n_struct)
            q_star = tf.concat([qt, rt], axis=-1)
        return q_star

    def get_config(self) -> dict:
        """
        Get config dict for serialization
        Returns: dict
        """
        d = super().get_config()
        d.update({"units": self.units, "num_steps": self.num_steps, "field": self.field})
        return d


@register
class WeightedReadout(ReadOut):
    """
    Perform a weighted average of the readout field. Weights are learnable
    from this layer
    """

    def __init__(self, neurons: List[int], activation="swish", field: str = "atoms", **kwargs):
        """

        Args:
            neurons (list): list of number of neurons in each layer
            activation (str): activation function
            field (str): the field to perform the readout, choose from
                "atoms" or "bonds"
            **kwargs:
        """
        super().__init__(**kwargs)
        n_layer = len(neurons)
        self.mlp = MLP(neurons=neurons, activations=[activation] * n_layer)
        neurons = neurons[:-1] + [1]  # type: ignore
        acts = [activation] * (n_layer - 1) + ["sigmoid"]
        self.weight = MLP(neurons=neurons, activations=acts)
        self.neurons = neurons
        self.activation = activation
        self.field = field

    def call(self, graph: List, **kwargs) -> tf.Tensor:
        """
        Args:
            graph (list): list repr of a MaterialGraph
            **kwargs:
        Returns:
        """
        field = graph[getattr(Index, self.field.upper())]
        n_field = graph[getattr(Index, f"n_{self.field}".upper())]

        updated_field = self.mlp(field)
        weights = self.weight(field)
        indices = get_segment_indices_from_n(n_field)
        n_structs = tf.shape(n_field)[0]
        factor = unsorted_segment_fraction(weights[:, 0], indices, num_segments=n_structs)
        return tf.math.segment_sum(factor[:, None] * updated_field, indices)

    def get_config(self):
        """
        Get config dict for serialization
        Returns: dict
        """
        config = super().get_config()
        config.update(
            {
                "neurons": self.neurons,
                "activation": self.activation,
                "field": self.field,
            }
        )
        return config



"""Aggregate classes. Aggregating happens when bond attributes """
from typing import Callable, List, Union

import tensorflow as tf

from m3gnet.graph import Index
from m3gnet.utils import get_segment_indices_from_n, register

from ._core import METHOD_MAPPING


@register
class AtomReduceState(tf.keras.layers.Layer):
    """
    Reduce atom attributes to states via sum or mean
    """

    def __init__(self, method: Union[Callable, str] = "mean", **kwargs):
        """

        Args:
            method (str or Callable): the method for the aggregation,
                choose from ["sum", "prod", "max", "min", "mean"]
                default to "mean"
            **kwargs:
        """
        super().__init__(**kwargs)
        self.method = method
        if isinstance(method, str):
            if method not in METHOD_MAPPING:
                raise ValueError("Unrecognized method name")
            method_func = METHOD_MAPPING[method]
        else:
            method_func = method
        self.method_func: Callable = method_func

    def call(self, graph: List, **kwargs) -> tf.Tensor:
        """
        Main layer logic
        Args:
            graph (list): input graph list representation
            **kwargs:
        Returns:
        """
        atom_segment = get_segment_indices_from_n(graph[Index.N_ATOMS])
        return self.method_func(
            graph[Index.ATOMS],
            atom_segment,
            num_segments=tf.shape(graph[Index.STATES])[0],
        )

    def get_config(self) -> dict:
        """
        Get the configuration dict
        Returns:

        """
        config = super().get_config()
        config.update({"method": self.method})
        return config



"""
M3GNet potential trainer
"""
from typing import List, Optional

import tensorflow as tf
from ase import Atoms
from pymatgen.core import Molecule, Structure

from m3gnet.callbacks import ManualStop
from m3gnet.graph import Index, MaterialGraph, MaterialGraphBatchEnergyForceStress
from m3gnet.layers import AtomRef
from m3gnet.models import Potential


class PotentialTrainer:
    """
    Trainer for M3GNet potential
    """

    def __init__(self, potential: Potential, optimizer: tf.keras.optimizers.Optimizer):
        """
        Args:
            potential (M3GNet): a M3GNet model
            optimizer (tf.keras.optimizers.Optimizer): a keras optimizer
        """
        self.potential = potential
        self.optimizer = optimizer

    def train(
        self,
        graphs_or_structures: List,
        energies: List,
        forces: List,
        stresses: Optional[List] = None,
        validation_graphs_or_structures: List = None,
        val_energies: List = None,
        val_forces: List = None,
        val_stresses: List = None,
        loss: tf.keras.losses.Loss = tf.keras.losses.MSE,
        force_loss_ratio: float = 1,
        stress_loss_ratio: float = 0.1,
        batch_size: int = 32,
        epochs: int = 1000,
        callbacks: List = None,
        save_checkpoint: bool = True,
        early_stop_patience: int = 200,
        verbose: int = 1,
        fit_per_element_offset: bool = False,
    ):
        """
        Args:
            graphs_or_structures (list): a list of MaterialGraph or structures
            energies (list): list of train energies
            forces (list): list of train forces
            stresses (list): list of train stresses
            validation_graphs_or_structures (list): optional list of validation
                graphs or structures
            val_energies (list): list of val energies
            val_forces (list): list of val forces
            val_stresses (list): list of val stresses
            loss (tf.keras.losses.Loss): loss object
            force_loss_ratio (float): the ratio of forces in loss
            stress_loss_ratio (float): the ratio of stresses in loss
            train_metrics (list): list of train metrics
            val_metrics (list): list of validation metrics
            val_monitor (str): field to monitor during validation, e.g.,
                "val_mae", "val_acc" or "val_auc"
            batch_size (int): batch size of combining graphs
            epochs (int): epochs for training the data
            callbacks (list): list of callback functions
            save_checkpoint (bool): whether to save model check point
            early_stop_patience (int): patience for early stop
            verbose (bool): whether to show model training progress
            fit_per_element_offset (bool): whether to train an element-wise
                offset, e.g., elemental energies etc. If trained, such energy
                will be summed to the neural network predictions.

        Returns: None
        """

        if isinstance(graphs_or_structures[0], MaterialGraph):
            graphs = graphs_or_structures
        elif isinstance(graphs_or_structures[0], (Structure, Molecule, Atoms)):
            graphs = [self.potential.model.graph_converter(i) for i in graphs_or_structures]
        else:
            raise ValueError("Graph types not recognized")

        if fit_per_element_offset:
            ar = AtomRef(max_z=self.potential.model.n_atom_types + 1)
            ar.fit(graphs, energies)
            self.potential.model.set_element_refs(ar.property_per_element)

        mgb = MaterialGraphBatchEnergyForceStress(
            graphs,
            energies=energies,
            forces=forces,
            stresses=stresses,
            batch_size=batch_size,
        )

        if validation_graphs_or_structures is not None and val_energies is not None:
            has_validation = True
            if isinstance(validation_graphs_or_structures[0], MaterialGraph):
                validation_graphs = validation_graphs_or_structures
            elif isinstance(validation_graphs_or_structures[0], (Structure, Molecule, Atoms)):
                validation_graphs = [self.potential.model.graph_converter(i) for i in validation_graphs_or_structures]
            else:
                raise ValueError("Graph types not recognized")

            mgb_val = MaterialGraphBatchEnergyForceStress(
                validation_graphs,
                energies=val_energies,
                forces=val_forces,
                stresses=val_stresses,
                batch_size=batch_size,
                shuffle=False,
            )
        else:
            has_validation = False

        has_stress = stresses is not None

        def _flat_loss(x, y):
            return loss(tf.reshape(x, (-1,)), tf.reshape(y, (-1,)))

        def _mae(x, y):
            x = tf.reshape(x, tf.shape(y))
            return tf.reduce_mean(tf.math.abs(x - y))

        def _loss(target_batch, graph_pred_batch, n_atoms):
            n_atoms_temp = tf.cast(n_atoms, dtype=target_batch[0].dtype)
            e_target = target_batch[0] / n_atoms_temp
            e_target = e_target[:, None]
            e_pred = graph_pred_batch[0] / n_atoms_temp[:, None]
            e_loss = _flat_loss(e_target, e_pred)
            f_loss = _flat_loss(target_batch[1], graph_pred_batch[1])
            e_metric = _mae(e_target, e_pred)
            f_metric = _mae(target_batch[1], graph_pred_batch[1])

            s_loss = 0
            s_metric = 0
            if has_stress:
                s_loss = _flat_loss(target_batch[2], graph_pred_batch[2])
                s_metric = _mae(target_batch[2], graph_pred_batch[2])
            return (
                e_loss + force_loss_ratio * f_loss + stress_loss_ratio * s_loss,
                e_metric,
                f_metric,
                s_metric,
            )

        if callbacks is None:
            callbacks = [tf.keras.callbacks.History()]

        if verbose:
            pbar = tf.keras.callbacks.ProgbarLogger(count_mode="steps")
            pbar.set_params({"verbose": verbose, "epochs": epochs})
            callbacks.append(pbar)

        callbacks.append(ManualStop())
        if has_validation and save_checkpoint:
            name_temp = "callbacks/{epoch:05d}-{val_MAE:.6f}-" "{val_MAE(E):.6f}-{val_MAE(F):.6f}"
            if has_stress:
                name_temp += "-{val_MAE(S):.6f}"
            callbacks.append(
                tf.keras.callbacks.ModelCheckpoint(
                    filepath=name_temp,
                    monitor="val_MAE",
                    save_weights_only=True,
                    save_best_only=True,
                    mode="min",
                )
            )

        if early_stop_patience:
            callbacks.append(tf.keras.callbacks.EarlyStopping(monitor="val_MAE", patience=200))

        callback_list = tf.keras.callbacks.CallbackList(callbacks)
        callback_list.on_train_begin()
        callback_list.set_model(self.potential.model)

        @tf.function(experimental_relax_shapes=True)
        def train_one_step(potential, graph_list, target_list):
            with tf.GradientTape() as tape:
                if has_stress:
                    pred_list = potential.get_efs_tensor(graph_list, include_stresses=True)
                else:
                    pred_list = potential.get_ef_tensor(graph_list)
                loss_val, emae, fmae, smae = _loss(target_list, pred_list, graph_list[Index.N_ATOMS])
            grads = tape.gradient(loss_val, potential.model.trainable_variables)
            return loss_val, grads, pred_list, emae, fmae, smae

        for epoch in range(epochs):
            callback_list.on_epoch_begin(epoch=epoch, logs={"epoch": epoch})
            epoch_loss_avg = tf.keras.metrics.Mean()
            emae_avg = tf.keras.metrics.Mean()
            fmae_avg = tf.keras.metrics.Mean()
            smae_avg = tf.keras.metrics.Mean()
            for batch_index, batch in enumerate(mgb):
                callback_list.on_batch_begin(batch=batch_index)
                graph_batch, target_batch = batch
                lossval, grads, pred_list, emae, fmae, smae = train_one_step(
                    self.potential, graph_batch.as_tf().as_list(), target_batch
                )

                self.optimizer.apply_gradients(zip(grads, self.potential.trainable_variables))
                epoch_loss_avg.update_state(lossval)
                emae_avg.update_state(emae)
                fmae_avg.update_state(fmae)
                smae_avg.update_state(smae)
                logs = {
                    "loss": epoch_loss_avg.result().numpy(),
                    "MAE(E)": emae_avg.result().numpy(),
                    "MAE(F)": fmae_avg.result().numpy(),
                    "MAE(S)": smae_avg.result().numpy(),
                }
                callback_list.on_batch_end(batch=batch_index, logs=logs)

            epoch_logs = {
                "loss": epoch_loss_avg.result().numpy(),
                "MAE(E)": emae_avg.result().numpy(),
                "MAE(F)": fmae_avg.result().numpy(),
                "MAE(S)": smae_avg.result().numpy(),
            }

            epoch_loss_avg = tf.keras.metrics.Mean()
            emae_avg = tf.keras.metrics.Mean()
            fmae_avg = tf.keras.metrics.Mean()
            smae_avg = tf.keras.metrics.Mean()

            for batch_index, batch in enumerate(mgb_val):
                graph_batch, target_batch = batch
                lossval, emae, fmae, smae = _loss(
                    target_batch,
                    self.potential.get_efs_tensor(graph_batch.as_tf().as_list(), True),
                    graph_batch.n_atoms,
                )
                epoch_loss_avg.update_state(lossval)
                emae_avg.update_state(emae)
                fmae_avg.update_state(fmae)
                smae_avg.update_state(smae)

            epoch_logs.update(
                **{
                    "val_MAE": emae_avg.result().numpy()
                    + force_loss_ratio * fmae_avg.result().numpy()
                    + stress_loss_ratio * smae_avg.result().numpy(),
                    "val_MAE(E)": emae_avg.result().numpy(),
                    "val_MAE(F)": fmae_avg.result().numpy(),
                    "val_MAE(S)": smae_avg.result().numpy(),
                }
            )

            callback_list.on_epoch_end(epoch=epoch, logs=epoch_logs)
            mgb.on_epoch_end()

            if getattr(self.potential.model, "stop_training", False):
                break



"""
Dynamics calculations using M3GNet
"""

import pickle
import contextlib
import sys
import io
from typing import Optional, Union

import numpy as np
from ase import Atoms, units
from ase.calculators.calculator import Calculator, all_changes
from ase.constraints import ExpCellFilter
from ase.io import Trajectory
from ase.md.nptberendsen import NPTBerendsen
from ase.md.nvtberendsen import NVTBerendsen
from ase.optimize.bfgs import BFGS
from ase.optimize.bfgslinesearch import BFGSLineSearch
from ase.optimize.fire import FIRE
from ase.optimize.lbfgs import LBFGS, LBFGSLineSearch
from ase.optimize.mdmin import MDMin
from ase.optimize.optimize import Optimizer
from ase.optimize.sciopt import SciPyFminBFGS, SciPyFminCG
from pymatgen.core import Molecule, Structure
from pymatgen.io.ase import AseAtomsAdaptor

from ._base import Potential
from ._m3gnet import M3GNet

OPTIMIZERS = {
    "FIRE": FIRE,
    "BFGS": BFGS,
    "LBFGS": LBFGS,
    "LBFGSLineSearch": LBFGSLineSearch,
    "MDMin": MDMin,
    "SciPyFminCG": SciPyFminCG,
    "SciPyFminBFGS": SciPyFminBFGS,
    "BFGSLineSearch": BFGSLineSearch,
}


class M3GNetCalculator(Calculator):
    """
    M3GNet calculator based on ase Calculator
    """

    implemented_properties = ["energy", "free_energy", "forces", "stress"]

    def __init__(self, potential: Potential, compute_stress: bool = True, stress_weight: float = 1.0, **kwargs):
        """

        Args:
            potential (Potential): m3gnet.models.Potential
            compute_stress (bool): whether to calculate the stress
            stress_weight (float): the stress weight.
            **kwargs:
        """
        super().__init__(**kwargs)
        self.potential = potential
        self.compute_stress = compute_stress
        self.stress_weight = stress_weight

    def calculate(
        self,
        atoms: Optional[Atoms] = None,
        properties: Optional[list] = None,
        system_changes: Optional[list] = None,
    ):
        """
        Args:
            atoms (ase.Atoms): ase Atoms object
            properties (list): list of properties to calculate
            system_changes (list): monitor which properties of atoms were
                changed for new calculation. If not, the previous calculation
                results will be loaded.
        Returns:

        """
        properties = properties or ["energy"]
        system_changes = system_changes or all_changes
        super().calculate(atoms=atoms, properties=properties, system_changes=system_changes)

        graph = self.potential.graph_converter(atoms)
        graph_list = graph.as_tf().as_list()
        results = self.potential.get_efs_tensor(graph_list, include_stresses=self.compute_stress)
        self.results.update(
            energy=results[0].numpy().ravel(),
            free_energy=results[0].numpy().ravel(),
            forces=results[1].numpy(),
        )
        if self.compute_stress:
            self.results.update(stress=results[2].numpy()[0] * self.stress_weight)


class Relaxer:
    """
    Relaxer is a class for structural relaxation
    """

    def __init__(
        self,
        potential: Optional[Union[Potential, str]] = None,
        optimizer: Union[Optimizer, str] = "FIRE",
        relax_cell: bool = True,
        stress_weight: float = 0.01,
    ):
        """

        Args:
            potential (Optional[Union[Potential, str]]): a potential,
                a str path to a saved model or a short name for saved model
                that comes with M3GNet distribution
            optimizer (str or ase Optimizer): the optimization algorithm.
                Defaults to "FIRE"
            relax_cell (bool): whether to relax the lattice cell
            stress_weight (float): the stress weight for relaxation
        """
        if isinstance(potential, str):
            potential = Potential(M3GNet.load(potential))
        if potential is None:
            potential = Potential(M3GNet.load())

        if isinstance(optimizer, str):
            optimizer_obj = OPTIMIZERS.get(optimizer, None)
        elif optimizer is None:
            raise ValueError("Optimizer cannot be None")
        else:
            optimizer_obj = optimizer

        self.opt_class: Optimizer = optimizer_obj
        self.calculator = M3GNetCalculator(potential=potential, stress_weight=stress_weight)
        self.relax_cell = relax_cell
        self.potential = potential
        self.ase_adaptor = AseAtomsAdaptor()

    def relax(
        self,
        atoms: Atoms,
        fmax: float = 0.1,
        steps: int = 500,
        traj_file: str = None,
        interval=1,
        verbose=False,
        **kwargs,
    ):
        """

        Args:
            atoms (Atoms): the atoms for relaxation
            fmax (float): total force tolerance for relaxation convergence.
                Here fmax is a sum of force and stress forces
            steps (int): max number of steps for relaxation
            traj_file (str): the trajectory file for saving
            interval (int): the step interval for saving the trajectories
            **kwargs:
        Returns:
        """
        if isinstance(atoms, (Structure, Molecule)):
            atoms = self.ase_adaptor.get_atoms(atoms)
        atoms.set_calculator(self.calculator)
        stream = sys.stdout if verbose else io.StringIO()
        with contextlib.redirect_stdout(stream):
            obs = TrajectoryObserver(atoms)
            if self.relax_cell:
                atoms = ExpCellFilter(atoms)
            optimizer = self.opt_class(atoms, **kwargs)
            optimizer.attach(obs, interval=interval)
            optimizer.run(fmax=fmax, steps=steps)
            obs()
        if traj_file is not None:
            obs.save(traj_file)
        if isinstance(atoms, ExpCellFilter):
            atoms = atoms.atoms

        return {
            "final_structure": self.ase_adaptor.get_structure(atoms),
            "trajectory": obs,
        }


class TrajectoryObserver:
    """
    Trajectory observer is a hook in the relaxation process that saves the
    intermediate structures
    """

    def __init__(self, atoms: Atoms):
        """
        Args:
            atoms (Atoms): the structure to observe
        """
        self.atoms = atoms
        self.energies: list[float] = []
        self.forces: list[np.ndarray] = []
        self.stresses: list[np.ndarray] = []
        self.atom_positions: list[np.ndarray] = []
        self.cells: list[np.ndarray] = []

    def __call__(self):
        """
        The logic for saving the properties of an Atoms during the relaxation
        Returns:
        """
        self.energies.append(self.compute_energy())
        self.forces.append(self.atoms.get_forces())
        self.stresses.append(self.atoms.get_stress())
        self.atom_positions.append(self.atoms.get_positions())
        self.cells.append(self.atoms.get_cell()[:])

    def compute_energy(self) -> float:
        """
        calculate the energy, here we just use the potential energy
        Returns:
        """
        energy = self.atoms.get_potential_energy()
        return energy

    def save(self, filename: str):
        """
        Save the trajectory to file
        Args:
            filename (str): filename to save the trajectory
        Returns:
        """
        with open(filename, "wb") as f:
            pickle.dump(
                {
                    "energy": self.energies,
                    "forces": self.forces,
                    "stresses": self.stresses,
                    "atom_positions": self.atom_positions,
                    "cell": self.cells,
                    "atomic_number": self.atoms.get_atomic_numbers(),
                },
                f,
            )


class MolecularDynamics:
    """
    Molecular dynamics class
    """

    def __init__(
        self,
        atoms: Atoms,
        potential: Optional[Union[Potential, str]] = "MP-2021.2.8-EFS",
        ensemble: str = "nvt",
        temperature: int = 300,
        timestep: float = 1.0,
        pressure: float = 1.01325 * units.bar,
        taut: Optional[float] = None,
        taup: Optional[float] = None,
        compressibility_au: Optional[float] = None,
        trajectory: Optional[Union[str, Trajectory]] = None,
        logfile: Optional[str] = None,
        loginterval: int = 1,
        append_trajectory: bool = False,
    ):
        """

        Args:
            atoms (Atoms): atoms to run the MD
            potential (Potential): potential for calculating the energy, force,
                stress of the atoms
            ensemble (str): choose from 'nvt' or 'npt'. NPT is not tested,
                use with extra caution
            temperature (float): temperature for MD simulation, in K
            timestep (float): time step in fs
            pressure (float): pressure in eV/A^3
            taut (float): time constant for Berendsen temperature coupling
            taup (float): time constant for pressure coupling
            compressibility_au (float): compressibility of the material in A^3/eV
            trajectory (str or Trajectory): Attach trajectory object
            logfile (str): open this file for recording MD outputs
            loginterval (int): write to log file every interval steps
            append_trajectory (bool): Whether to append to prev trajectory
        """

        if isinstance(potential, str):
            potential = Potential(M3GNet.load(potential))

        if isinstance(atoms, (Structure, Molecule)):
            atoms = AseAtomsAdaptor().get_atoms(atoms)
        self.atoms = atoms
        self.atoms.set_calculator(M3GNetCalculator(potential=potential))

        if taut is None:
            taut = 100 * timestep * units.fs
        if taup is None:
            taup = 1000 * timestep * units.fs

        if ensemble.lower() == "nvt":
            self.dyn = NVTBerendsen(
                self.atoms,
                timestep * units.fs,
                temperature_K=temperature,
                taut=taut,
                trajectory=trajectory,
                logfile=logfile,
                loginterval=loginterval,
                append_trajectory=append_trajectory,
            )

        elif ensemble.lower() == "npt":
            self.dyn = NPTBerendsen(
                self.atoms,
                timestep * units.fs,
                temperature_K=temperature,
                pressure_au=pressure,
                taut=taut,
                taup=taup,
                compressibility_au=compressibility_au,
                trajectory=trajectory,
                logfile=logfile,
                loginterval=loginterval,
                append_trajectory=append_trajectory,
            )
        else:
            raise ValueError("Ensemble not supported")

        self.trajectory = trajectory
        self.logfile = logfile
        self.loginterval = loginterval
        self.timestep = timestep

    def run(self, steps: int):
        """
        Thin wrapper of ase MD run
        Args:
            steps (int): number of MD steps
        Returns:

        """
        self.dyn.run(steps)

    def set_atoms(self, atoms: Atoms):
        """
        Set new atoms to run MD
        Args:
            atoms (Atoms): new atoms for running MD

        Returns:

        """
        calculator = self.atoms.calc
        self.atoms = atoms
        self.dyn.atoms = atoms
        self.dyn.atoms.set_calculator(calculator)
